/*
 *  kernel_interface.c
 *
 *  Copyright (C) 1999 Standard & Western Software
 *  http://www.sw-soft.com
 *  This software is distributed AS IS without
 *  warranty of any kind.
 *
 */
 
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <KernelExport.h>
#include <time.h>
#include "fsproto.h"
#include <malloc.h>

#include "ntfs.h"

/*  Start of fundamental (read-only) required functions */
static int	fs_mount(nspace_id nsid, const char *device, ulong flags,
						void *parms, size_t len, void **data, vnode_id *vnid);
static int	fs_unmount(void *_ns);
static int	fs_walk(void *_ns, void *_base, const char *file,
						char **newpath, vnode_id *vnid);
static int	fs_read_vnode(void *_ns, vnode_id vnid, char r, void **node);
static int	fs_write_vnode(void *_ns, void *_node, char r);
static int	fs_rstat(void *_ns, void *_node, struct stat *st);
static int	fs_open(void *_ns, void *_node, int omode, void **cookie);
static int	fs_read(void *_ns, void *_node, void *cookie, off_t pos,
						void *buf, size_t *len);
static int	fs_write(void *ns, void *node, void *cookie, off_t pos,
					const void *buf, size_t *len);
static int	fs_free_cookie(void *ns, void *node, void *cookie);
static int	fs_close(void *ns, void *node, void *cookie);
static int	fs_access(void *_ns, void *_node, int mode);
static int	fs_opendir(void* _ns, void* _node, void** cookie);
static int	fs_readdir(void *_ns, void *_node, void *cookie,
					long *num, struct dirent *buf, size_t bufsize);
static int	fs_rewinddir(void *_ns, void *_node, void *cookie);
static int	fs_closedir(void *_ns, void *_node, void *cookie);
static int	fs_free_dircookie(void *_ns, void *_node, void *cookie);
static int	fs_rfsstat(void *_ns, struct fs_info *);
static int 	fs_readlink(void *_ns, void *_node, char *buf, size_t *bufsize);



/* End of fundamental (read-only) required functions. */


/* vnode_ops struct. Fill this in to tell the kernel how to call
	functions in your driver.
*/

vnode_ops fs_entry =  {
	&fs_read_vnode,						/* read_vnode func ptr */
	&fs_write_vnode,					/* write_vnode func ptr */
	NULL, 								/* remove_vnode func ptr */
	NULL,								/* secure_vnode func ptr */
	&fs_walk,							/* walk func ptr */
	&fs_access,							/* access func ptr */
	NULL, 								/* create func ptr */
	NULL, 								/* mkdir func ptr */
	NULL,
	NULL,
	NULL,
	NULL, 								/* unlink func ptr */
	NULL, 								/* rmdir func ptr */
	&fs_readlink,						/* readlink func ptr */
	&fs_opendir,						/* opendir func ptr */
	&fs_closedir,						/* closedir func ptr */
	&fs_free_dircookie,					/* free_dircookie func ptr */
	&fs_rewinddir,						/* rewinddir func ptr */
	&fs_readdir,						/* readdir func ptr */
	&fs_open,							/* open file func ptr */
	&fs_close,							/* close file func ptr */
	&fs_free_cookie,					/* free cookie func ptr */
	&fs_read,							/* read file func ptr */
	&fs_write, 							/* write file func ptr */
	NULL, /* readv */
	NULL, /* writev */
	NULL,								/* ioctl func ptr */
	NULL,								/* setflags file func ptr */
	&fs_rstat,							/* rstat func ptr */
	NULL, 								/* wstat func ptr */
	NULL,
	NULL, 								/* initialize func ptr */
	&fs_mount,							/* mount func ptr */
	&fs_unmount,						/* unmount func ptr */
	NULL,								/* sync func ptr */
	&fs_rfsstat,						/* rfsstat func ptr */
	NULL,								/* wfsstat func ptr */
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

int32	api_version = B_CUR_FS_API_VERSION;

static char *gFSName = "NTFS";

thread_id gSerializeThread;


static int fs_mount( nspace_id nsid, const char *device, ulong flags, void *parms,
					 size_t len, void **data, vnode_id *vnid )
{
	/*
	Kernel passes in nspace_id, (representing a disk or partition?)
	and a string representing the device (eg, "/dev/scsi/disk/030/raw)
	Flags will be used for things like specifying read-only mounting.
	parms is parameters passed in as switches from the mount command, 
	and len is the length of the otions. data is a pointer to a 
	driver-specific struct that should be allocated in this routine. 
	It will then be passed back in by the kernel to a number of the other 
	fs driver functions. vnid should also be passed back to the kernel, 
	representing the vnode id of the root vnode.
	*/
//	#pragma unused(flags)
//	#pragma unused(parms)
//	#pragma unused (len)
	int 	result = B_NO_ERROR;
	nspace	*vol;
	vnode	*newNode = NULL;

//	set_dprintf_enabled(true);
//	dprintf("fs_mount - ENTER\n");
//	dprintf("%s: nsid %d device %s\n",gFSName, nsid,device);

	result = NTFSMount(device, O_RDONLY, &vol);
	if (result == B_NO_ERROR)
	{
		*vnid = NTFS_ROOTNODE_ID;
		*data = (void*)vol;
		vol->id = nsid;
		
		newNode = (vnode*)ntfs_calloc( sizeof(vnode) );
		if ( newNode == NULL )
			result = ENOMEM;
		else
		{
			ntfs_init_inode( &newNode->ntnode, vol->ntvol, *vnid );
			result = new_vnode( nsid, *vnid, (void*)newNode );
			if ( result != B_NO_ERROR )
			{
				ntfs_free( vol );
				result = EINVAL;
			}
			else 
				result = B_NO_ERROR;
		}
	}

//	dprintf("fs_mount - EXIT, result code is %s\n", strerror(result));
	return result;
}

static int fs_unmount( void *_ns )
{
	int		result = B_NO_ERROR;
	nspace	*ns = (nspace*)_ns;

//	dprintf("fs_unmount - ENTER\n");
//	dprintf("fs_unmount - closing volume\n");
	result = close( ns->ntvol->u.fd );
//	dprintf("fs_unmount - freeing volume data\n");

	ntfs_release_volume( ns->ntvol );
	ntfs_free( ns );

//	dprintf("fs_unmount - EXIT, result is %s\n", strerror(result));
//	set_dprintf_enabled(false);
	return result;
}

static int fs_rfsstat( void *_ns, struct fs_info * fss )
{
	nspace	*ns = (nspace*)_ns;
	
//	dprintf("fs_rfsstat - ENTER\n");
	
	fss->dev = ns->ntvol->u.fd;
	fss->dev = ns->id;
	fss->root = NTFS_ROOTNODE_ID;
	fss->flags = B_FS_IS_PERSISTENT | B_FS_IS_READONLY;
	fss->block_size = ns->ntvol->clustersize;
	fss->io_size = 65536;
	fss->total_blocks = ntfs_get_volumesize( ns->ntvol );
	fss->free_blocks = ns->free_cluster_count;
	strncpy( fss->device_name, ns->devicePath, sizeof(fss->device_name) );
	strncpy( fss->volume_name, ns->volumeLabel, sizeof(fss->volume_name) );
	strcpy( fss->fsh_name, "NTFS" );

//	dprintf("fs_rfsstat - EXIT\n");
	return B_NO_ERROR;
}

/*
    fs_walk - the walk function just "walks" through a directory looking for
	the specified file. When you find it, call get_vnode on its vnid to init
	it for the kernel.
*/

static int fs_walk( void *_ns, void *base, const char *file,
					char **newpath, vnode_id *vnid )
{
	/* Starting at the base, find file in the subdir, and return path
		string and vnode id of file. */
	nspace	*ns = (nspace*)_ns;
	vnode	*baseNode = (vnode*)base;
	vnode	*newNode = NULL;
	int		result = ENOENT;

//	dprintf("fs_walk - ENTER : find for %s at basenode=%d, parent=%d\n",
//			file,baseNode->ntnode.i_number,baseNode->parent_number);

	if ( !strcmp( file, "." ) )
	{
		*vnid = baseNode->ntnode.i_number;
		if ( get_vnode( ns->id, *vnid, (void**)&newNode ) != 0 )
			result = EINVAL;
		else
			result = B_NO_ERROR;
	}
	else if ( !strcmp( file, ".." ) && baseNode->ntnode.i_number != NTFS_ROOTNODE_ID )
	{
		*vnid = baseNode->parent_number;
		if ( get_vnode( ns->id, *vnid, (void**)&newNode ) != 0 )
			result = EINVAL;
		else
			result = B_NO_ERROR;
	}
	else
	{
		*vnid = ntfs_find_file( &baseNode->ntnode, file );
		if ( *vnid == -1 )
			result = EINVAL;
		else
			if ( get_vnode( ns->id, *vnid, (void**)&newNode ) != 0 )
				result = EINVAL;
		else
		{
			newNode->parent_number = baseNode->ntnode.i_number;
			result = B_NO_ERROR;
		}
	}

//	dprintf("fs_walk - EXIT, result is %s, vnid is %Lu\n", strerror(result), *vnid);
	return result;
}

/*
   fs_read_vnode - Using vnode id, read in vnode information into fs-specific struct,
					and return it in node. the reenter flag tells you if this function
					is being called via some other fs routine, so that things like 
					double-locking can be avoided.
*/
static int fs_read_vnode( void *_ns, vnode_id vnid, char reenter, void **node )
{
//	#pragma unused (reenter)
	nspace	*ns = (nspace*)_ns;
	int		result = B_NO_ERROR;
	vnode	*newNode = NULL;
	
//	dprintf("fs_read_vnode - ENTER, vnid is %d\n", vnid);
	newNode = (vnode*)ntfs_calloc( sizeof(vnode) );
	if ( newNode != NULL )
	{
		ntfs_init_inode( &newNode->ntnode, ns->ntvol, vnid );
//		dprintf("fs_read_vnode - init result is %s\n", strerror(result));
		*node = (void*)newNode;
	}
	else
		result = ENOMEM;
//	dprintf("fs_read_vnode - EXIT, result is %s\n", strerror(result));
	return result;
}

static int fs_write_vnode( void *ns, void *_node, char reenter )
{
//	#pragma unused (ns)
//	#pragma unused (reenter)
	int 	result = B_NO_ERROR;
	vnode	*node = (vnode*)_node;
	
//	dprintf("fs_write_vnode - ENTER (0x%x)\n", node);
	if ( node != NULL )
	{
		ntfs_clear_inode( &node->ntnode );
		ntfs_free( node );
	}
//	dprintf("fs_write_vnode - EXIT\n");
	return result;
}

static int fs_rstat( void *_ns, void *_node, struct stat *st )
{
	nspace	*ns = (nspace*)_ns;
	vnode	*node = (vnode*)_node;
	int 	result = B_NO_ERROR;
	time_t	time;
	ntfs_io	io;
	char	buffer[0x30];
	ntfs_attribute *at_data = NULL;
	
//	dprintf("fs_rstat - ENTER: node=%d\n",node->ntnode.i_number);
	st->st_dev = ns->id;
	st->st_ino = node->ntnode.i_number;

	st->st_nlink = NTFS_GETU16( node->ntnode.attr + 0x12 );
	st->st_uid = 0;
	st->st_gid = 0;
	st->st_blksize = 65536;
	if ( NTFS_GETU16(node->ntnode.attr+0x16) & 2 )
		st->st_mode = __S_IFDIR | 0555;
	else
		st->st_mode = __S_IFREG | 0444;

	io.param = (void*)buffer;
	io.size = 0x30;
	io.fn_put = ntfs_put;
	io.fn_get = ntfs_get;
	if ( ntfs_read_attr( &node->ntnode, ns->ntvol->at_standard_information, NULL, 0, &io ) )
		return EINVAL;
	st->st_crtime = ntfs_ntutc2unixutc( NTFS_GETU64(buffer) );
	st->st_ctime = st->st_mtime = ntfs_ntutc2unixutc( NTFS_GETU64(buffer+0x8) );
	st->st_atime = ntfs_ntutc2unixutc( NTFS_GETU64(buffer+0x18) );

	at_data = ntfs_find_attr( &node->ntnode, ns->ntvol->at_data, NULL );
	st->st_size = at_data ? at_data->size : 2048;
//	dprintf("fs_rstat - EXIT, result is %s\n", strerror(result));
	return result;
}

/*
   fs_open - Create a vnode cookie, if necessary, to use when
 				reading/writing a file
*/
static int fs_open( void *_ns, void *_node, int omode, void **cookie )
{
//	#pragma unused (_ns)
//	#pragma unused (cookie)
	int		result = B_NO_ERROR;
	vnode	*node = (vnode*)_node;
	
//	dprintf("fs_open - ENTER\n");
//	dprintf("node=%d\n",node->i_number);

//	if ( NTFS_GETU16(node->attr+0x16) & 2 )
//		result = EISDIR;

	// No cookie info needed, just return.

//	dprintf("fs_open - EXIT\n");
	return result;
}

/*
   fs_read
   Read a file specified by node, using information in cookie
   and at offset specified by pos. read len bytes into buffer buf.
*/
static int fs_read( void *_ns, void *_node, void *cookie, off_t pos,
					void *buf, size_t *len )
{
//	#pragma unused (cookie)
	nspace	*ns = (nspace*)_ns;
	vnode	*node = (vnode*)_node;
	int		result = B_NO_ERROR;
	ntfs_io	io;

	io.do_read = 1;
	io.param = (void*)buf;
	io.size = *len;
	io.fn_put = ntfs_put;
	io.fn_get = ntfs_get;
	if ( ntfs_read_attr( &node->ntnode, ns->ntvol->at_data, NULL, pos, &io ) )
		result = EINVAL;

	*len = io.size;
//	dprintf("fs_read - EXIT, result is %s\n", strerror(result));
	return result;
}

/*
   fs_write
   In this read-only driver the function does nothing.
*/
static int fs_write( void *ns, void *node, void *cookie, off_t pos,
					 const void *buf, size_t *len )
{
//	#pragma unused (ns)
//	#pragma unused (node)
//	#pragma unused (cookie)
//	#pragma unused (pos)
//	#pragma unused (buf)
//	#pragma unused (len)
//	dprintf("fs_write - ENTER\n");
//	dprintf("fs_write - EXIT\n");
	return B_NO_ERROR;
}

/*
   fs_close - Do whatever is necessary to close a file, EXCEPT for freeing
				the cookie!
*/
static int fs_close( void *ns, void *node, void *cookie )
{
//	#pragma unused (ns)
//	#pragma unused (node)
//	#pragma unused (cookie)
//	dprintf("fs_close - ENTER\n");
//	dprintf("fs_close - EXIT\n");
	return B_NO_ERROR;
}

static int fs_free_cookie( void *ns, void *node, void *cookie )
{
//	#pragma unused (ns)
//	#pragma unused (node)
//	#pragma unused (cookie)

	// We don't allocate file cookies, so we do nothing here.
//	dprintf("fs_free_cookie - ENTER\n");
//	dprintf("fs_free_cookie - EXIT\n");
	return B_NO_ERROR;
}

/*
	fs_access - checks permissions for access.

	ns 		- global, fs-specific struct for device
	node	- node to check permissions for
	mode 	- requested permissions on node.
*/
static int fs_access( void *ns, void *node, int mode )
{
//	#pragma unused (ns)
//	#pragma unused (node)
//	#pragma unused (mode)

//	dprintf("fs_access - ENTER\n");
//	dprintf("fs_access - EXIT\n");
	return B_NO_ERROR;
}

static int fs_readlink(void *_ns, void *_node, char *buf, size_t *bufsize)
{
//	#pragma unused(_ns)

	int		result = EINVAL;
	vnode*	node = (vnode*)_node;
	size_t	length = 0;
	
/*
	if (S_ISLNK(node->attr.stat[FS_DATA_FORMAT].st_mode))
	{
		length = strlen(node->attr.slName);
		if (length > *bufsize)
		{
			memcpy(buf, node->attr.slName, length);
		}
		else
		{
			memcpy(buf, node->attr.slName, length);
		}
		result = B_NO_ERROR;
	}
*/
	*bufsize = length;
	return result;
}

/*
	fs_opendir - creates fs-specific "cookie" struct that keeps track of where
					you are at in reading through directory entries in fs_readdir.

	ns 		- global, fs-specific struct for device
	node	- directory to open
	cookie	- allocate fs-specific cookie and return here to
			  maintain dirent position info for reads.	
*/
static int fs_opendir( void *_ns, void *_node, void **_cookie )
{
	vnode		*node = (vnode*)_node;
	int			result = B_NO_ERROR;
	dircookie	*cookie = (dircookie*)ntfs_calloc( sizeof(dircookie) );

//	dprintf("fs_opendir - ENTER, node=%ld\n", node->ntnode.i_number); 

	if ( !(NTFS_GETU16(node->ntnode.attr+0x16) & 2) )
		result = EMFILE;

	if ( cookie != NULL )
	{
		cookie->walk.type = BY_POSITION;
		cookie->walk.dir = &node->ntnode;
		cookie->walk.u.pos = 0;
		cookie->first = 1;
		cookie->state = 0;
		*_cookie = (void*)cookie;
	}
	else
		result = ENOMEM;
//	dprintf("fs_opendir - EXIT\n");
	return result;
}

/*
   fs_readdir - read 1 or more dirents, keep state in cookie, return
					0 when no more entries.
*/
static int fs_readdir( void *_ns, void *_node, void *_cookie, long *num, 
					   struct dirent *buf, size_t bufsize )
{
	int 		result = B_NO_ERROR;
	nspace		*ns = (nspace*)_ns;
	vnode		*node = (vnode*)_node;
	dircookie	*cookie = (dircookie*)_cookie;
	
//	dprintf("fs_readdir - ENTER\n");
	result = NTFSReadDirEnt( ns, cookie, buf, bufsize );
	if ( result == B_NO_ERROR )
		*num = 1;
	else
		*num = 0;
	
//	When you get to the end, don't return an error, just return
//	a zero in *num.
	if ( result == ENOENT )
		result = B_NO_ERROR;
//	dprintf("fs_readdir - EXIT, result is %s\n", strerror(result));
	return result;
}
			
/*
	fs_rewinddir - set cookie to represent beginning of directory, so
					later fs_readdir calls start at beginning.

	ns 		- global, fs-specific struct for device
	node	- directory to rewind
	cookie	- current cookie for directory.	
*/
static int fs_rewinddir( void *ns, void *_node, void *_cookie )
{
//	#pragma unused (ns)
//	#pragma unused (node)

	int			result = EINVAL;
	dircookie	*cookie = (dircookie*)_cookie;

//	dprintf("fs_rewinddir - ENTER\n");
	if ( cookie != NULL )
	{
		cookie->walk.u.pos = 0;
		cookie->first = 1;
		cookie->state = 0;
		result = B_NO_ERROR;
	}
//	dprintf("fs_rewinddir - EXIT, result is %s\n", strerror(result));
	return result;
}

/*
	fs_closedir - Do whatever you need to to close a directory (sometimes
					nothing), but DON'T free the cookie!

	ns 		- global, fs-specific struct for device
	node	- directory to close
	cookie	- current cookie for directory.	
*/
static int fs_closedir( void *ns, void *node, void *cookie )
{
//	#pragma unused (ns)
//	#pragma unused (node)
//	#pragma unused (cookie)

//	dprintf("fs_closedir - ENTER\n");
//	dprintf("fs_closedir - EXIT\n");
	return B_NO_ERROR;
}

/*
	fs_free_dircookie - Free the fs-specific cookie struct

	ns 		- global, fs-specific struct for device
	node	- directory related to cookie
	cookie	- current cookie for directory, to free.	
*/
static int fs_free_dircookie( void *ns, void *node, void *cookie )
{
//	#pragma unused (ns)
//	#pragma unused (node)

//	dprintf("fs_free_dircookie - ENTER\n");
	if ( cookie != NULL )
	{
		ntfs_free( cookie );
	}
//	dprintf("fs_free_dircookie - EXIT\n");
	return B_NO_ERROR;
}
