/*
 *  ntfs.c
 *
 *  Copyright (C) 1999 Standard & Western Software
 *  http://www.sw-soft.com
 *  This software is distributed AS IS without
 *  warranty of any kind.
 *
 */
 
#include <malloc.h>
#include <SupportDefs.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <Drivers.h>
#include <OS.h>
#include <KernelExport.h> 
#include <string.h>
#include <dirent.h>
#include <time.h>
#include <malloc.h>

#include "ntfs.h"
#include "libntfs/nttools.h"

/*
	path: 		path to device (eg, /dev/disk/scsi/030/raw)
	partition:	partition number on device ????
	flags:		currently unused
*/

int NTFSMount( char *path, int flags, nspace **newVol )
{
//	#pragma unused (flags)

	nspace*		vol = NULL;
	int 		result = B_NO_ERROR;
	char		cluster0[512];
	
//	dprintf("NTFSMount - ENTER\n");
	vol = (nspace*)ntfs_calloc( sizeof(nspace) );
	if ( vol == NULL )
		result = ENOMEM;
	else
	{
		vol->ntvol = ntfs_open_volume( path, 0, 1, 0 );
		if ( vol->ntvol )
		{
			strncpy( vol->devicePath, path, 127 );
			vol->ntvol->ngt = ngt_nt;
			vol->ntvol->nct = nct_utf8;
			vol->free_cluster_count = ntfs_get_free_cluster_count( vol->ntvol->bitmap );
			result = NTFSReadVolumeLabel( vol );
			if ( result != B_NO_ERROR )
			{
				ntfs_free( vol );
				vol = NULL;
			}
		}
		else
		{
			ntfs_free( vol );
			vol = NULL;
			result = EINVAL;
		}
	}
//	dprintf("NTFSMount - EXIT, result %s, returning 0x%x\n", strerror(result), vol);
	*newVol = vol;
	return result;
}

/* Reads the volume label */
int NTFSReadVolumeLabel( nspace *ns )
{
	int result = B_NO_ERROR;
	ntfs_inode *vol_inode = (ntfs_inode*)ntfs_calloc( sizeof(ntfs_inode) );
	if ( vol_inode == NULL )
		result = ENOMEM;
	else
		if ( ntfs_init_inode( vol_inode, ns->ntvol, FILE_VOLUME ) )
			return EINVAL;
		else
		{
			ntfs_attribute *at_volname = NULL;
			ntfs_io io;
			char buffer[127];
			at_volname = ntfs_find_attr( vol_inode, 0x60, NULL );
			if ( at_volname == NULL )
				result = EINVAL;
			else
			{
				io.param = (void*)buffer;
				io.size = at_volname->size < 127 ? at_volname->size : 127;
				io.fn_put = ntfs_put;
				io.fn_get = ntfs_get;
				if ( ntfs_read_attr( vol_inode, 0x60, NULL, 0, &io ) )
					result = EINVAL;
				else
				{
					char *name;
					int name_len;
					if ( result=ntfs_encodeuni( ns->ntvol, (ntfs_u16*)buffer, at_volname->size/2,
							 &name, &name_len ) )
						result = EINVAL;
					else
					{
						strncpy( ns->volumeLabel, name, 127 );
						ntfs_free( name );
					}
				}
			}
		}
	if ( vol_inode != NULL )
	{
		ntfs_clear_inode( vol_inode );
		ntfs_free( vol_inode );
	}
	return result;
} 

/* Reads in a single directory entry and fills in the values in the
	dirent struct. Uses the cookie to keep track of the current block
	and position withing the block. Also uses the cookie to determine when
	it has reached the end of the directory file.
	
	NOTE: If your file sytem seems to work ok from the command line, but
	the tracker doesn't seem to like it, check what you do here closely;
	in particular, if the d_ino in the stat struct isn't correct, the tracker
	will not display the entry.
*/
int NTFSReadDirEnt( nspace *ns, dircookie *cookie,
					struct dirent *buf, size_t bufsize )
{
	int	result = B_NO_ERROR;
	char item[1000];
	char *name;
	int name_len;
	int flags;

//	dprintf("NTFSReadDirEnt - ENTER for dir=%ld\n",cookie->walk.dir->i_number);
	buf->d_ino = 0;
	buf->d_reclen = 0;
	buf->d_name[0] = '\0';
	cookie->walk.result = item;
	for ( ;; )
	{
		// generate special extra files for non-root directories

		if ( cookie->state == 0 && cookie->walk.dir->i_number != NTFS_ROOTNODE_ID )
		{
			buf->d_reclen = 1;
			strcpy( buf->d_name, "." );
			cookie->state = 1;
			break;
		}
		if ( cookie->state == 1 && cookie->walk.dir->i_number != NTFS_ROOTNODE_ID )
		{
			buf->d_reclen = 2;
			strcpy( buf->d_name, ".." );
			cookie->state = 2;
			break;
		}

		// get the next entry
		if ( !ntfs_getdir_byposition(&cookie->walk) )
		{
			result = ENOENT;
			break;
		}
		// on the first iteration, we won't get any data
		if ( cookie->first )
		{
			cookie->first = 0;
			continue;
		}
		flags = NTFS_GETU8( item+0x51 );
		if ( flags == 2 )
			continue;
		if ( ntfs_encodeuni( ns->ntvol, (ntfs_u16*)(item+0x52), NTFS_GETU8(item+0x50),
							 &name, &name_len ) )
		{
			result = EINVAL;
			break;
		}
		if ( name[0] == '$' && cookie->walk.dir->i_number == NTFS_ROOTNODE_ID )
		{
			free( name );
			continue;
		}

		buf->d_ino = ntfs_find_file( cookie->walk.dir, name );
		buf->d_reclen = name_len;
		strncpy( buf->d_name, name, name_len+1 );
		ntfs_free( name );
		break;
	}
//	dprintf("NTFSReadDirEnt - EXIT, result is %s, vnid is %Lu\n", strerror(result),buf->d_ino);
	return result;
}
